#!/usr/bin/python3

# Wrapper around cloud-localds and libvirt

# Copyright (C) 2012-9 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import base64
import errno
import functools
import io
import itertools
import os
import platform
import re
import shutil
import signal
import string
import subprocess
import sys
import tempfile
import uuid
import yaml

import libvirt
from lxml import etree
from lxml.builder import E, ElementMaker

import uvtool.libvirt
from uvtool.libvirt import LIBVIRT_METADATA_XMLNS
import uvtool.libvirt.simplestreams
import uvtool.ssh
import uvtool.wait


ARCH = platform.machine()
DEFAULT_TEMPLATE = '/usr/share/uvtool/libvirt/template.xml'

DEFAULT_REMOTE_WAIT_SCRIPT = '/usr/share/uvtool/libvirt/remote-wait.sh'
POOL_NAME = 'uvtool'


class CLIError(Exception):
    """An error that should be reflected back to the CLI user."""
    pass


class InsecureError(RuntimeError):
    """An insecure operation is required and the user did not permit it by
    using --insecure."""
    pass


def get_template_path(target_arch):
    if uvtool.libvirt.is_emulation_target(target_arch):
        if target_arch == 'armhf':
            return '/usr/share/uvtool/libvirt/template-emu-armhf.xml'

    if target_arch == 'aarch64':
        return '/usr/share/uvtool/libvirt/template-aarch64.xml'
    elif target_arch == 'ppc64el':
        return '/usr/share/uvtool/libvirt/template-ppc64le.xml'
    elif target_arch == 's390x':
        return '/usr/share/uvtool/libvirt/template-s390x.xml'
    elif target_arch == 'amd64' or target_arch == 'i386':
        return DEFAULT_TEMPLATE

    print(
        "Warning: unknown target architecture '%s' (host %s) using defaults"
        % (target_arch, ARCH), file=sys.stderr
    )
    return DEFAULT_TEMPLATE


# From: http://www.chiark.greenend.org.uk/ucgi/~cjwatson/blosxom/2009-07-02-python-sigpipe.html
def subprocess_setup():
    # Python installs a SIGPIPE handler by default. This is usually not what
    # non-Python subprocesses expect.
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def run_script_once_arg_to_config(arg, unique_id):
    with open(arg, 'rb') as f:
        script = f.read()
    encoded_script = base64.b64encode(script).decode('ascii')
    return [
        'cloud-init-per',
        'once',
        unique_id,
        'sh', '-c',
        (
            'f=$(mktemp --tmpdir %s-XXXXXXXXXX) && ' +
            'echo "%s" | base64 -d > "$f" && ' +
            'chmod 700 "$f" && ' +
            '"$f" && ' +
            'rm "$f"'
        ) % (unique_id, encoded_script)
    ]


def run_script_once_args_to_config(args):
    return [
        run_script_once_arg_to_config(arg, 'uvt-kvm-%s' % i)
        for i, arg in enumerate(args)
    ]


def get_ssh_agent_public_keys():
    """Read public keys from an agent, if available, or return None."""
    try:
        with open("/dev/null") as fpnull:
            output = subprocess.check_output(
                ['ssh-add', '-L'],
                stderr=fpnull
            ).decode()
    except subprocess.CalledProcessError:
        return None

    output = output.strip()
    if output:
        return output.splitlines()
    else:
        return None


def read_ssh_public_key_file(filename):
    """Read public keys from a file, or return None."""

    if filename is None:
        filename = os.path.join(os.environ['HOME'], '.ssh', 'id_rsa.pub')

    try:
        f = open(filename, 'r')
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise
        return None, filename
    else:
        with f:
            return f.read().strip().splitlines(), filename


def get_ssh_authorized_keys(filename):
    # If the user hasn't explicitly specified a file source, then try
    # the agent first.
    if filename is None:
        agent_keys = get_ssh_agent_public_keys()
        if agent_keys:
            return agent_keys
        # Fall back to reading the public key file.

    # Read the public key file if one is present.
    file_keys, filename_used = read_ssh_public_key_file(filename)
    if file_keys:
        return file_keys
    else:
        print(
            "Warning: %s not found; instance will be started "
            "with no ssh access by default." % repr(filename_used),
            file=sys.stderr,
        )
        return []


def create_default_user_data(fobj, args, ssh_host_keys=None):
    """Write some sensible default cloud-init user-data to the given file
    object.

    """

    ssh_authorized_keys = list(
        get_ssh_authorized_keys(args.ssh_public_key_file),
    )

    data = {
        'hostname': args.hostname,
        'manage_etc_hosts': 'localhost',
        'ssh_keys': uvtool.ssh.generate_ssh_host_keys()[0],
    }

    if ssh_host_keys:
        data['ssh_keys'] = ssh_host_keys

    if ssh_authorized_keys:
        data['ssh_authorized_keys'] = ssh_authorized_keys

    if args.password:
        data['password'] = args.password
        data['chpasswd'] = {'expire': False}
        data['ssh_pwauth'] = True

    if args.run_script_once:
        data['runcmd'] = run_script_once_args_to_config(args.run_script_once)

    if args.packages:
        data['packages'] = list(
            itertools.chain(*[p.split(',') for p in args.packages])
        )

    fobj.write(b"#cloud-config\n")
    fobj.write(yaml.dump(data).encode())


def create_default_meta_data(fobj, args):
    data = {
        'instance-id': str(uuid.uuid1()),
    }
    fobj.write(yaml.dump(data).encode())


def create_ds_image(
        temp_dir,
        hostname,
        user_data_fobj,
        meta_data_fobj,
        network_config_fobj=None
    ):
    """Create a file called ds.img inside temp_dir that contains a useful
    cloud-init data source.

    Other temporary files created in temp_dir are currently metadata and
    userdata and network-config and can be safely deleted.

    """

    with open(os.path.join(temp_dir, 'userdata'), 'wb') as f:
        f.write(user_data_fobj.read())
    with open(os.path.join(temp_dir, 'metadata'), 'wb') as f:
        f.write(meta_data_fobj.read())
    if network_config_fobj:
        with open(os.path.join(temp_dir, 'network-config'), 'wb') as f:
            f.write(network_config_fobj.read())

    args = ['cloud-localds', '--disk-format=qcow2']
    if network_config_fobj:
        args.extend(['--network-config', 'network-config'])
    args.extend(['ds.img', 'userdata', 'metadata'])
    subprocess.check_call(args, cwd=temp_dir)


def create_ds_volume(
        new_volume_name,
        hostname,
        user_data_fobj,
        meta_data_fobj,
        network_config_fobj=None
    ):
    """Create a new libvirt cloud-init datasource volume."""

    temp_dir = tempfile.mkdtemp(prefix='uvt-kvm-')
    try:
        create_ds_image(
            temp_dir=temp_dir,
            hostname=hostname,
            user_data_fobj=user_data_fobj,
            meta_data_fobj=meta_data_fobj,
            network_config_fobj=network_config_fobj,
        )
        with open(os.path.join(temp_dir, 'ds.img'), 'rb') as f:
            return uvtool.libvirt.create_volume_from_fobj(
                new_volume_name, f, image_type='qcow2', pool_name=POOL_NAME)
    finally:
        shutil.rmtree(temp_dir)


def create_new_volume(new_volume_name, size=2):
    """Create a new libvirt volume with provided name and size."""

    temp_dir = tempfile.mkdtemp(prefix='uvt-kvm-')
    try:
        fname = os.path.join(temp_dir, 'disk.qcow')
        output = subprocess.check_output(
            ['qemu-img', 'create', '-f', 'qcow2', fname, "%dG" % size])
        with open(fname, 'rb') as f:
            return uvtool.libvirt.create_volume_from_fobj(
                new_volume_name, f, image_type='qcow2', pool_name=POOL_NAME)
    finally:
        shutil.rmtree(temp_dir)


def create_cow_volume(backing_volume_name, new_volume_name, new_volume_size,
        conn=None):

    if conn is None:
        conn = libvirt.open('qemu:///system')

    pool = conn.storagePoolLookupByName(POOL_NAME)
    try:
        backing_vol = pool.storageVolLookupByName(backing_volume_name)
    except libvirt.libvirtError:
        raise RuntimeError("Cannot find volume %s" % backing_volume_name)

    return create_cow_volume_by_path(
        backing_volume_path=backing_vol.path(),
        new_volume_name=new_volume_name,
        new_volume_size=new_volume_size,
        conn=conn
    )

def create_cow_volume_by_path(backing_volume_path, new_volume_name,
        new_volume_size, conn=None):
    """Create a new libvirt qcow2 volume backed by an existing volume path."""

    if conn is None:
        conn = libvirt.open('qemu:///system')

    pool = conn.storagePoolLookupByName(POOL_NAME)

    new_vol = E.volume(
        E.name(new_volume_name),
        E.allocation('0'),
        E.capacity(str(new_volume_size), unit='G'),
        E.target(E.format(type='qcow2')),
        E.backingStore(
            E.path(backing_volume_path),
            E.format(type='qcow2'),
            )
        )
    return pool.createXML(etree.tostring(new_vol, encoding=str), 0)


def get_boot_item_name(hostname, item_name):
    """fuse hostname and item_name to make up a unique filename
    This works always, if the file already exists or not.
    The name will be used as identifier in the storage pool"""
    return hostname + "-" + item_name


def get_boot_item_path(item_name, conn=None):
    """boot items (when extracted) are placed in the pool
    this helper gets that path (if one already exists) and
    returns None if it doesn't."""
    if conn is None:
        conn = libvirt.open('qemu:///system')

    pool = conn.storagePoolLookupByName(POOL_NAME)
    try:
        path = pool.storageVolLookupByName(item_name).path()
    except libvirt.libvirtError:
        return None

    return path


def remove_boot_item(hostname, item, conn=None):
    """Remove boot item if it exists in the pool"""
    item_name = get_boot_item_name(hostname, item)

    if conn is None:
        conn = libvirt.open('qemu:///system')

    pool = conn.storagePoolLookupByName(POOL_NAME)
    try:
        item_vol = pool.storageVolLookupByName(item_name)
        item_vol.delete(0)
    except libvirt.libvirtError:
        pass


def get_boot_item(backing_volume_name, hostname, item, conn=None):
    """Extract the boot item from /boot of the backing image
    and return a path to the extracted item.
    If the file already exists use that - this allows users to place
    arbitrary files on their own for using different kernels or just the
    inability to run as root."""

    if conn is None:
        conn = libvirt.open('qemu:///system')

    item_name = get_boot_item_name(hostname, item)
    item_path = get_boot_item_path(item_name, conn)
    if item_path is not None and os.path.exists(item_path):
        return item_path

    if not (os.geteuid() == 0):
        raise CLIError("emulation needs root to extract kernel/initrd")

    pool = conn.storagePoolLookupByName(POOL_NAME)
    try:
        backing_path = pool.storageVolLookupByName(backing_volume_name).path()
    except libvirt.libvirtError:
        raise RuntimeError("Cannot find volume %s" % backing_volume_name)

    temp_mic_dir = tempfile.mkdtemp(prefix='uvt-kvm-mic')
    source = os.path.join(temp_mic_dir, "boot", item)
    temp_extract_dir = tempfile.mkdtemp(prefix='uvt-kvm-extract')
    temp_target = os.path.join(temp_extract_dir, item)
    args = ['mount-image-callback']
    args.extend(
        [backing_path, '--mountpoint', temp_mic_dir, '--', 'cp', '-L',
         source, temp_target]
    )
    try:
        subprocess.check_call(args)
        with open(temp_target, 'rb') as f:
            item_vol = uvtool.libvirt.create_volume_from_fobj_with_size(
                item_name,
                fobj=f,
                fobj_size=os.fstat(f.fileno()).st_size,
                image_type="raw",
                pool_name=POOL_NAME)
    except subprocess.CalledProcessError:
        print("error: extract %s from image %s" % (item, backing_path))
    finally:
        shutil.rmtree(temp_mic_dir)
        shutil.rmtree(temp_extract_dir)

    return item_vol.path()


def compose_domain_xml(name, volumes, template_path, cpu=1, memory=512,
        unsafe_caching=False, log_console_output=False, host_passthrough=False,
        machine_type=None, bridge=None, mac=None, ssh_known_hosts=None,
        kernel=None, initrd=None):
    tree = etree.parse(template_path)
    domain = tree.getroot()
    assert domain.tag == 'domain'

    etree.strip_elements(domain, 'name')
    etree.SubElement(domain, 'name').text = name

    etree.strip_elements(domain, 'vcpu')
    etree.SubElement(domain, 'vcpu').text = str(cpu)

    etree.strip_elements(domain, 'currentMemory')
    etree.SubElement(domain, 'currentMemory').text = str(memory * 1024)

    etree.strip_elements(domain, 'memory')
    etree.SubElement(domain, 'memory').text = str(memory * 1024)

    devices = domain.find('devices')

    etree.strip_elements(devices, 'disk')
    for num, vol in enumerate(volumes):
        disk_device = "vd%s" % string.ascii_letters[num]
        disk_format_type = (
            etree.fromstring(vol.XMLDesc(0)).
            find('target').
            find('format').
            get('type')
            )
        if unsafe_caching:
            disk_driver = E.driver(
                name='qemu', type=disk_format_type, cache='unsafe')
        else:
            disk_driver = E.driver(name='qemu', type=disk_format_type)
        devices.append(
            E.disk(
                disk_driver,
                E.source(file=vol.path()),
                E.target(dev=disk_device),
                type='file',
                device='disk',
                )
            )

    if bridge:
        etree.strip_elements(devices, 'interface')
        if mac:
            devices.append(E.interface(
                             E.mac(address=mac),
                             E.source(bridge=bridge),
                             E.model(type='virtio'),
                             type='bridge'),
                          )
        else:
            devices.append(E.interface(
                             E.source(bridge=bridge),
                             E.model(type='virtio'),
                             type='bridge'),
                          )
    else:
        if mac:
            etree.strip_elements(devices, 'interface')
            devices.append(E.interface(
                                 E.mac(address=mac),
                                 E.source(network='default'),
                                 E.model(type='virtio'),
                                 type='network'),
                          )

    if log_console_output:
        if ARCH == 's390x':
            raise CLIError("logging guest console output is currently"
                           "not supported on s390x.")
        else:
            print(
                "Warning: logging guest console output introduces a DoS " +
                "security problem on the host and should not be used in " +
                "production.",
                file=sys.stderr
            )
            etree.strip_elements(devices, 'serial')
            devices.append(E.serial(E.target(port='0'), type='stdio'))

    if host_passthrough:
        if ARCH == 'aarch64':
            print("Info: on aarch64 a host type cpu is the default",
                  file=sys.stderr)
        else:
            etree.SubElement(domain, 'cpu', mode='host-passthrough')

    if machine_type:
        domos = domain.find('os')
        domtype = domos.find('type')
        domtype.set('machine', machine_type)

    if initrd:
       domos = domain.find('os')
       etree.SubElement(domos, 'initrd').text = initrd

    if kernel:
       domos = domain.find('os')
       etree.SubElement(domos, 'kernel').text = kernel

    if ssh_known_hosts:
        metadata = domain.find('metadata')
        if metadata is None:
            metadata = E.metadata()
            domain.append(metadata)
        EX = ElementMaker(
            namespace=LIBVIRT_METADATA_XMLNS,
            nsmap={'uvt': LIBVIRT_METADATA_XMLNS}
        )
        metadata.append(EX.ssh_known_hosts(ssh_known_hosts))

    return etree.tostring(tree, encoding=str)


def get_base_image(filters):
    result = list(uvtool.libvirt.simplestreams.query(filters))
    if not result:
        raise CLIError(
            "no images found that match filters %s." % repr(filters))
    elif len(result) != 1:
        raise CLIError(
            "multiple images found that match filters %s." % repr(filters))
    return result[0]


def create(
        hostname,
        filters,
        user_data_fobj,
        meta_data_fobj,
        network_config_fobj,
        template_path,
        memory=512,
        cpu=1,
        disk=2,
        unsafe_caching=False,
        log_console_output=False,
        host_passthrough=False,
        bridge=None,
        mac=None,
        backing_image_file=None,
        start=True,
        ssh_known_hosts=None,
        ephemeral_disks=None,
        machine_type=None,
        target_arch=None
    ):
    if backing_image_file is None:
        base_volume_name = get_base_image(filters)
    if ephemeral_disks is None:
        ephemeral_disks = []
    undo_volume_creation = []
    try:
        # cow image names must end in ".qcow" so that the current Apparmor
        # profile for /usr/lib/libvirt/virt-aa-helper is able to read them,
        # determine their backing volumes, and generate a dynamic libvirt
        # profile that permits reading the backing volume. Once our pool
        # directory is added to the virt-aa-helper profile, this requirement
        # can be dropped.

        if backing_image_file:
            main_vol = create_cow_volume_by_path(
                backing_image_file, "%s.qcow" % hostname, disk)
        else:
            main_vol = create_cow_volume(
                base_volume_name, "%s.qcow" % hostname, disk)
        undo_volume_creation.append(main_vol)

        ds_vol = create_ds_volume(
            "%s-ds.qcow" % hostname,
            hostname,
            user_data_fobj,
            meta_data_fobj,
            network_config_fobj,
        )
        undo_volume_creation.append(ds_vol)

        volumes = [main_vol, ds_vol]
        for num, ephem_size in enumerate(ephemeral_disks):
            vol = create_new_volume(
                "%s-ephem-%02d.qcow" % (hostname, num), ephem_size)
            undo_volume_creation.append(vol)
            volumes.append(vol)

        if not machine_type and ARCH == 'x86_64':
            if target_arch and target_arch == 'amd64':
                machine_type = get_latest_q35_type()

        if target_arch and uvtool.libvirt.is_emulation_target(target_arch):
            kernel = get_boot_item(base_volume_name, hostname, "vmlinuz")
            initrd = get_boot_item(base_volume_name, hostname, "initrd.img")
        else:
            kernel = None
            initrd = None

        xml = compose_domain_xml(
            hostname, volumes=volumes,
            bridge=bridge,
            mac=mac,
            cpu=cpu,
            log_console_output=log_console_output,
            host_passthrough=host_passthrough,
            memory=memory,
            template_path=template_path,
            unsafe_caching=unsafe_caching,
            ssh_known_hosts=ssh_known_hosts,
            machine_type=machine_type,
            kernel=kernel,
            initrd=initrd,
        )

        conn = libvirt.open('qemu:///system')
        domain = conn.defineXML(xml)
        if start:
            try:
                domain.create()
            except:
                if ARCH == 'aarch64':
                    # aarch runs with nvram per our default template, flag
                    # needed to be able to remove those on undefine
                    domain.undefineFlags(libvirt.VIR_DOMAIN_UNDEFINE_NVRAM)
                else:
                    domain.undefine()
                raise
    except:
        for vol in undo_volume_creation:
            vol.delete(0)
        raise


def delete_domain_volumes(conn, domain):
    """Delete all volumes associated with a domain.

    :param conn: libvirt connection object
    :param domain: libvirt domain object

    """
    domain_xml = etree.fromstring(domain.XMLDesc(0))
    assert domain_xml.tag == 'domain'
    for disk in domain_xml.find('devices').iter('disk'):
        disk_file = disk.find('source').get('file')
        vol = conn.storageVolLookupByKey(disk_file)
        vol.delete(0)


def destroy(hostname):
    conn = libvirt.open('qemu:///system')
    try:
        domain = conn.lookupByName(hostname)
    except libvirt.libvirtError as e:
        if e.get_error_code() == libvirt.VIR_ERR_NO_DOMAIN:
            raise CLIError("domain %s not found." % repr(hostname))
        else:
            raise
    state = domain.state(0)[0]
    if state != libvirt.VIR_DOMAIN_SHUTOFF:
        domain.destroy()

    delete_domain_volumes(conn, domain)
    remove_boot_item(hostname, "vmlinuz", conn)
    remove_boot_item(hostname, "initrd.img", conn)


    if ARCH == 'aarch64':
        # aarch runs with nvram per our default template, flag
        # needed to be able to remove those on undefine
        domain.undefineFlags(libvirt.VIR_DOMAIN_UNDEFINE_NVRAM)
    else:
        domain.undefine()


def get_latest_q35_type():
    """Return the latest supported q35 type by the local qemu

    That isn't as simple as using distro.os_release_info()['ubuntu_codename']
    since that would not accound for older or newer (UCA) qemu versions.

    Returns None if it fails to find a proper type, which lets libvirt choose
    a default.
    """
    try:
        output = subprocess.check_output(
            ['qemu-system-x86_64', '-M', '?'], close_fds=True
        ).decode()
        p = re.compile(r'^pc-q35-[a-z]+ +Ubuntu [0-9]+\.[0-9]+', re.M)
        q35lines = p.findall(output)
        q35types = [line.split() for line in q35lines]
        q35types = sorted(q35types, key=lambda x: x[2], reverse=True)
        machine_type = q35types[0][0]
        if not machine_type.strip():
            raise ValueError
    except (IndexError, ValueError, subprocess.CalledProcessError):
        return None

    return machine_type


def get_lts_series():
    output = subprocess.check_output(['distro-info', '--lts'], close_fds=True)
    return output.strip().decode()


def apply_default_fobj(args, key, create_default_data_fn):
    """Return the fobj specified, creating it if required.

    If the "key" attribute inside args is not None, then just return it.
    Otherwise, construct a new temporary fobj, populate it with the result of
    create_default_data_fn, and then return that.

    This is useful to apply default file objects to argparse.FileType command
    line parameters.

    """
    specified_fobj = getattr(args, key)
    if specified_fobj:
        return specified_fobj
    else:
        default_fobj = io.BytesIO()
        create_default_data_fn(default_fobj, args)
        default_fobj.seek(0)
        return default_fobj


def check_kvm_ok():
    try:
        process = subprocess.Popen(
            ['kvm-ok'], shell=False, stdout=subprocess.PIPE, close_fds=True)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        # Ignore if we can't find kvm-ok executable
        return True, None
    stdout, stderr = process.communicate()
    return (False, stdout) if process.returncode else (True, None)


def name_to_ips(name):
    macs = uvtool.libvirt.get_domain_macs(name)
    return [
        ip for ip
        in (uvtool.libvirt.mac_to_ip(mac) for mac in macs)
        if ip
    ]


def ssh(name, login_name, arguments, stdin=None, checked=False, sysexit=True,
        private_key_file=None, insecure=False):
    ips = name_to_ips(name)
    ip_count = len(ips)
    if not ip_count:
        raise CLIError(
            "no IP address found for libvirt machine %s. "
            "Has it had time to boot yet?\nTry: %s wait" %
                (repr(name), sys.argv[0]))
    elif ip_count > 1:
        raise CLIError(
            "multiple IPs detected for %s %s and are not supported." %
                (repr(name), repr(ips))
        )
    ip = ips[0]

    objects_to_close = []
    try:
        ssh_call = [
            'ssh',
        ]

        ssh_known_hosts = uvtool.libvirt.get_domain_ssh_known_hosts(
            name, prefix=('%s ' % ip)
        )
        if ssh_known_hosts:
            ssh_known_hosts_file = tempfile.NamedTemporaryFile(
                mode='w',
                prefix='uvt-kvm.known_hoststmp')
            objects_to_close.append(ssh_known_hosts_file)
            ssh_known_hosts_file.write(ssh_known_hosts)
            ssh_known_hosts_file.flush()
            ssh_call.extend(
                ['-o', 'UserKnownHostsFile=%s' % ssh_known_hosts_file.name]
            )
        else:
            if not insecure:
                raise InsecureError()
            ssh_call.extend([
                '-o', 'UserKnownHostsFile=/dev/null',
                '-o', 'StrictHostKeyChecking=no',
                '-o', 'CheckHostIP=no',
            ])

        if login_name:
            ssh_call.extend(['-l', login_name])
        if private_key_file:
            ssh_call.extend(['-i', private_key_file])
        ssh_call.append(ip)
        ssh_call.extend(arguments)

        call = subprocess.check_call if checked else subprocess.call

        result = call(
            ssh_call, preexec_fn=subprocess_setup, close_fds=True, stdin=stdin
        )

        if sysexit:
            sys.exit(result)

        return result
    finally:
        [x.close() for x in objects_to_close]


def main_create(parser, args):
    if args.user_data and args.password:
        parser.error("--password cannot be used with --user-data.")
    if args.password:
        print(
            "Warning: using --password from the command line is " +
                "not secure and should be used for debugging only.",
            file=sys.stderr
        )

    kvm_ok, is_kvm_ok_output = check_kvm_ok()
    if not kvm_ok:
        print(
            "KVM not available. kvm-ok returned:", is_kvm_ok_output,
            sep="\n", end="", file=sys.stderr
        )
        return

    ssh_host_keys, ssh_known_hosts = uvtool.ssh.generate_ssh_host_keys()

    user_data_fobj = apply_default_fobj(
        args, 'user_data', functools.partial(
            create_default_user_data,
            ssh_host_keys=ssh_host_keys
        )
    )
    meta_data_fobj = apply_default_fobj(
        args, 'meta_data', create_default_meta_data
    )

    if args.guest_arch:
        target_arch = args.guest_arch
    else:
        target_arch = uvtool.libvirt.platform_arch_to_image_arch(ARCH)

    template = get_template_path(target_arch)
    # an explicit template overrides general and guest-arch defaults
    if args.template:
        template = args.template

    if args.backing_image_file:
        abs_image_backing_file = os.path.abspath(args.backing_image_file)
    else:
        abs_image_backing_file = None
    create(
        hostname=args.hostname,
        filters=args.filters,
        user_data_fobj=user_data_fobj,
        meta_data_fobj=meta_data_fobj,
        network_config_fobj=args.network_config,
        backing_image_file=abs_image_backing_file,
        bridge=args.bridge,
        mac=args.mac,
        cpu=args.cpu,
        disk=args.disk,
        log_console_output=args.log_console_output,
        host_passthrough=args.host_passthrough,
        memory=args.memory,
        template_path=template,
        unsafe_caching=args.unsafe_caching,
        start=not args.no_start,
        ssh_known_hosts=ssh_known_hosts,
        ephemeral_disks=args.ephemeral_disks,
        machine_type=args.machine_type,
        target_arch=target_arch,
    )


def main_destroy(parser, args):
    for h in args.hostname:
        destroy(h)


def main_list(parser, args):
    # Hack for now. In time this should properly use the API and list
    # only instances created with this tool.
    subprocess.check_call('virsh -q list --all|awk \'{print $2}\'', shell=True)


def main_ip(parser, args):
    ips = name_to_ips(args.name)
    count = len(ips)
    if not count:
        raise CLIError(
            "no IP address found for libvirt machine %s." % repr(args.name))
    elif count > 1:
        print(
            "Warning: multiple IP address found for libvirt machine %s; " +
                "listing the first one only",
            file=sys.stderr
        )
    print(ips[0])


def main_ssh(parser, args, default_login_name='ubuntu'):
    if args.login_name:
        login_name = args.login_name
        name = args.name
    elif '@' in args.name:
        login_name, name = args.name.split('@', 1)
    else:
        login_name = default_login_name
        name = args.name

    try:
        return ssh(
            name, login_name, args.ssh_arguments, insecure=args.insecure)
    except InsecureError:
        raise CLIError(
            "ssh public host key not found. " +
                "Use --insecure iff you trust your network path to the guest."
        )


def main_wait_remote(parser, args):
    with open(args.remote_wait_script, 'rb') as wait_script:
        try:
            ssh(
                args.name,
                args.remote_wait_user,
                [
                    'env',
                    'UVTOOL_WAIT_INTERVAL=%s' % args.interval,
                    'UVTOOL_WAIT_TIMEOUT=%s' % args.timeout,
                    'sh',
                    '-'
                ],
                checked=True,
                stdin=wait_script,
                private_key_file=args.ssh_private_key_file,
                insecure=args.insecure,
            )
        except InsecureError:
            raise CLIError(
                "ssh public host key not found. Use "
                    "--insecure iff you trust your network path to the guest."
            )


def main_wait(parser, args):
    conn = libvirt.open('qemu:///system')
    domain = conn.lookupByName(args.name)
    state = domain.state(0)[0]
    if state != libvirt.VIR_DOMAIN_RUNNING:
        raise CLIError(
            "libvirt domain %s is not running." % repr(args.name))

    macs = list(uvtool.libvirt.get_domain_macs(args.name))
    if not macs:
        raise CLIError(
            "libvirt domain %s has no NIC MACs available." % repr(args.name))
    if len(macs) > 1:
        raise CLIError(
            "libvirt domain %s has more than one NIC defined."
                % repr(args.name)
        )
    mac = macs[0]
    if not uvtool.wait.wait_for_libvirt_dnsmasq_lease(
            mac, args.timeout):
        raise CLIError(
            "timed out waiting for dnsmasq lease for %s." % mac)
    host_ip = uvtool.libvirt.mac_to_ip(mac)
    if not uvtool.wait.wait_for_open_ssh_port(
            host_ip, args.interval, args.timeout):
        raise CLIError(
            "timed out waiting for ssh to open on %s." % host_ip)
    if not args.without_ssh:
        main_wait_remote(parser, args)


class DeveloperOptionAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.unsafe_caching = True
        namespace.log_console_output = True


def main(args):
    # Workaround for https://bugzilla.redhat.com/show_bug.cgi?id=1063766
    # (LP: #1228231)
    def noop(*args, **kwargs): pass
    libvirt.registerErrorHandler(noop, None)

    parser = argparse.ArgumentParser()
    # Workarounds needed here: set dest and required
    # https://bugs.python.org/issue9253#msg181855
    # https://stackoverflow.com/q/22990977/478206
    # https://stackoverflow.com/a/23354355/478206
    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = True
    create_subparser = subparsers.add_parser('create')
    create_subparser.set_defaults(func=main_create)
    create_subparser.add_argument(
        '--developer', '-d', nargs=0, action=DeveloperOptionAction)
    create_subparser.add_argument('--template', default=None)
    create_subparser.add_argument('--memory', default=512, type=int)
    create_subparser.add_argument('--cpu', default=1, type=int)
    create_subparser.add_argument('--disk', default=8, type=int)
    create_subparser.add_argument(
        '--ephemeral-disk', action='append', type=int, dest='ephemeral_disks',
        help='Add an empty disk of SIZE in GB', metavar='SIZE')
    create_subparser.add_argument('--bridge')
    create_subparser.add_argument('--mac')
    create_subparser.add_argument('--unsafe-caching', action='store_true')
    create_subparser.add_argument(
        '--user-data', type=argparse.FileType('rb'))
    create_subparser.add_argument(
        '--meta-data', type=argparse.FileType('rb'))
    create_subparser.add_argument(
        '--network-config', type=argparse.FileType('rb'))
    create_subparser.add_argument('--password')
    create_subparser.add_argument('--guest-arch',
        help='guest arch to select template, default is the host architecture')
    create_subparser.add_argument('--log-console-output', action='store_true')
    create_subparser.add_argument('--host-passthrough', action='store_true')
    create_subparser.add_argument('--machine-type')
    create_subparser.add_argument('--backing-image-file')
    create_subparser.add_argument('--run-script-once', action='append')
    create_subparser.add_argument('--ssh-public-key-file')
    create_subparser.add_argument('--packages', action='append')
    create_subparser.add_argument('--no-start', action='store_true', default=False)
    create_subparser.add_argument('hostname')
    create_subparser.add_argument(
        'filters', nargs='*', metavar='filter',
        default=["release=%s" % get_lts_series()],
    )
    destroy_subparser = subparsers.add_parser('destroy')
    destroy_subparser.set_defaults(func=main_destroy)
    destroy_subparser.add_argument('hostname', nargs='+')
    list_subparser = subparsers.add_parser('list')
    list_subparser.set_defaults(func=main_list)
    ip_subparser = subparsers.add_parser('ip')
    ip_subparser.set_defaults(func=main_ip)
    ip_subparser.add_argument('name')
    ssh_subparser = subparsers.add_parser('ssh')
    ssh_subparser.set_defaults(func=main_ssh)
    ssh_subparser.add_argument('--insecure', action='store_true')
    ssh_subparser.add_argument('--login-name', '-l')
    ssh_subparser.add_argument('name')
    ssh_subparser.add_argument('ssh_arguments', nargs='*')
    wait_subparser = subparsers.add_parser('wait')
    wait_subparser.set_defaults(func=main_wait)
    wait_subparser.add_argument('--timeout', type=float, default=120.0)
    wait_subparser.add_argument('--interval', type=float, default=8.0)
    wait_subparser.add_argument('--remote-wait-script',
        default=DEFAULT_REMOTE_WAIT_SCRIPT)
    wait_subparser.add_argument('--insecure', action='store_true')
    wait_subparser.add_argument('--remote-wait-user', default='ubuntu')
    wait_subparser.add_argument('--without-ssh', action='store_true')
    wait_subparser.add_argument('--ssh-private-key-file')
    wait_subparser.add_argument('name')
    args = parser.parse_args(args)
    args.func(parser, args)


def main_cli_wrapper(*args, **kwargs):
    try:
        main(*args, **kwargs)
    except CLIError as e:
        print(
            "%s: error: %s" % (os.path.basename(sys.argv[0]), e),
            file=sys.stderr
        )
        sys.exit(1)
    except libvirt.libvirtError as e:
        libvirt_message = e.get_error_message()
        print(
            "%s: error: libvirt: %s" % (
                os.path.basename(sys.argv[0]),
                libvirt_message
            ),
            file=sys.stderr
        )
        sys.exit(1)


if __name__ == '__main__':
    main_cli_wrapper(sys.argv[1:])



#!/usr/bin/python3

# Keep Ubuntu Cloud images synced to a local libvirt storage pool.

# Copyright (C) 2013-9 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This is written using Python 2 because libvirt bindings were not available
# for Python 3 at the time of writing.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import base64
import codecs
import collections
import errno
import json
import os
import subprocess

import libvirt

import simplestreams.filters
import simplestreams.mirrors
import simplestreams.util

import uvtool.libvirt

LIBVIRT_POOL_NAME = 'uvtool'
IMAGE_DIR = '/var/lib/uvtool/libvirt/images/' # must end in '/'; see use
METADATA_DIR = '/var/lib/uvtool/libvirt/metadata'
USEFUL_FIELD_NAMES = ['release', 'arch', 'label']


def mkdir_p(path):
    """Create path if it doesn't exist already"""
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


class Metadata(object):
    def __init__(self, metadata_dir=METADATA_DIR):
        self.metadata_dir = metadata_dir
        # Deliberately do not create metadata_dir as a side-effect here, since
        # this class is instantiated below when the module is loaded, and we
        # want to be able to test this module without the side-effect of it
        # affecting the system metadata directory.

    def _metadata_path(self, encoded_libvirt_image_name):
        return os.path.join(self.metadata_dir, encoded_libvirt_image_name)

    def __contains__(self, encoded_libvirt_image_name):
        return os.path.exists(self._metadata_path(encoded_libvirt_image_name))

    def __delitem__(self, encoded_libvirt_image_name):
        os.unlink(self._metadata_path(encoded_libvirt_image_name))

    def __setitem__(self, encoded_libvirt_image_name, metadata):
        mkdir_p(self.metadata_dir)
        with codecs.open(
                self._metadata_path(encoded_libvirt_image_name), 'wb',
                encoding='utf-8') as f:
            json.dump(metadata, f, indent=4)

    def __getitem__(self, encoded_libvirt_image_name):
        with codecs.open(
                self._metadata_path(encoded_libvirt_image_name), 'rb',
                encoding='utf-8') as f:
            return json.load(f)

    def keys(self):
        try:
            return os.listdir(self.metadata_dir)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            return []

    def items(self):
        return [(key, self[key]) for key in self.keys()]


pool_metadata = Metadata(METADATA_DIR)

BASE64_PREFIX = 'x-uvt-b64-'

def _encode_libvirt_pool_name(product_name, version_name):
    return BASE64_PREFIX + base64.b64encode(
        (' '.join([product_name, version_name])).encode(), b'-_'
    ).decode()


def _decode_libvirt_pool_name(encoded_pool_name):
    if not encoded_pool_name.startswith(BASE64_PREFIX):
        raise ValueError(
            "Volume name cannot be parsed for simplestreams identity: %s" %
            repr(encoded_pool_name)
        )
    return base64.b64decode(
        encoded_pool_name[len(BASE64_PREFIX):],
        altchars=b'-_'
    ).decode().split(None, 1)


def purge_pool(conn=None):
    '''Delete all volumes and metadata with prejudice.

    This removes images from the pool whether they are in use or not.

    '''
    # Remove all metadata first. If this is interrupted, then it just looks
    # like there are volumes waiting to be cleaned up.
    for key in pool_metadata.keys():
        del pool_metadata[key]

    # Remove actual volumes themselves
    if conn is None:
        conn = libvirt.open('qemu:///system')
    pool = uvtool.libvirt.get_libvirt_pool_object(conn, LIBVIRT_POOL_NAME)
    for volume_name in pool.listVolumes():
        volume = pool.storageVolLookupByName(volume_name)
        volume.delete(0)


def clean_extraneous_images():
    conn = libvirt.open('qemu:///system')
    pool = uvtool.libvirt.get_libvirt_pool_object(conn, LIBVIRT_POOL_NAME)
    encoded_libvirt_pool_names = uvtool.libvirt.volume_names_in_pool(
        LIBVIRT_POOL_NAME)
    volume_names_in_use = frozenset(
        uvtool.libvirt.get_all_domain_volume_names(
            filter_by_dir=IMAGE_DIR)
    )
    for encoded_libvirt_name in encoded_libvirt_pool_names:
        if (encoded_libvirt_name not in volume_names_in_use and
                not encoded_libvirt_name in pool_metadata):
            uvtool.libvirt.delete_volume_by_name(
                encoded_libvirt_name, pool_name=LIBVIRT_POOL_NAME)


def _load_products(path=None, content_id=None, clean=False):
    # If clean evaluates to True, then remove any metadata files for which
    # the corresponding volume is missing.
    def new_product():
        return {'versions': {}}
    products = collections.defaultdict(new_product)
    for encoded_libvirt_name_string, metadata in list(pool_metadata.items()):
        if not uvtool.libvirt.have_volume_by_name(
                encoded_libvirt_name_string, pool_name=LIBVIRT_POOL_NAME):
            if clean:
                del pool_metadata[encoded_libvirt_name_string]
            continue
        product, version = _decode_libvirt_pool_name(
            encoded_libvirt_name_string)
        assert(product == metadata['product_name'])
        assert(version == metadata['version_name'])
        products[product]['versions'][version] = {
            'items': { 'disk1.img': metadata }
        }
    return {'content_id': content_id, 'products': products}


class LibvirtQuery(simplestreams.mirrors.BasicMirrorWriter):
    def __init__(self, filters):
        super(LibvirtQuery, self).__init__()
        self.filters = filters
        self.result = []

    def load_products(self, path=None, content_id=None):
        return {'content_id': content_id, 'products': {}}

    def filter_item(self, data, src, target, pedigree):
        return simplestreams.filters.filter_item(
            self.filters, data, src, pedigree)

    def insert_item(self, data, src, target, pedigree, contentsource):
        product_name, version_name, item_name = pedigree
        self.result.append((product_name, version_name))


def query(filter_args):
    query = LibvirtQuery(simplestreams.filters.get_filters(filter_args))
    query.sync_products(None, src=_load_products())
    return (_encode_libvirt_pool_name(product_name, version_name)
        for product_name, version_name in query.result)

class LibvirtMirror(simplestreams.mirrors.BasicMirrorWriter):
    def __init__(self, filters, verbose=False):
        super(LibvirtMirror, self).__init__({'max_items': 1})
        self.filters = filters
        self.verbose = verbose

    def load_products(self, path=None, content_id=None):
        return _load_products(path=path, content_id=content_id, clean=True)

    def filter_index_entry(self, data, src, pedigree):
        return data['datatype'] == 'image-downloads'

    def filter_item(self, data, src, target, pedigree):
        return simplestreams.filters.filter_item(
            self.filters, data, src, pedigree)

    def insert_item(self, data, src, target, pedigree, contentsource):
        product_name, version_name, item_name = pedigree
        assert(item_name == 'disk1.img')
        if self.verbose:
            print("Adding: %s %s" % (product_name, version_name))
        encoded_libvirt_name = _encode_libvirt_pool_name(
            product_name, version_name)
        if not uvtool.libvirt.have_volume_by_name(
                encoded_libvirt_name, pool_name=LIBVIRT_POOL_NAME):
            uvtool.libvirt.create_volume_from_fobj(
                encoded_libvirt_name, contentsource, image_type='qcow2',
                pool_name=LIBVIRT_POOL_NAME
            )
        pool_metadata[encoded_libvirt_name] = (
            simplestreams.util.products_exdata(src, pedigree)
        )

    def remove_version(self, data, src, target, pedigree):
        product_name, version_name = pedigree
        if self.verbose:
            print("Removing: %s %s" % (product_name, version_name))
        encoded_libvirt_name = _encode_libvirt_pool_name(
            product_name, version_name)
        del pool_metadata[encoded_libvirt_name]


def main_sync(args):
    (mirror_url, initial_path) = simplestreams.util.path_from_mirror_url(
        args.mirror_url, args.path)

    def policy(content, path):
        if initial_path.endswith('sjson') and not args.no_authentication:
            return simplestreams.util.read_signed(
                content, keyring=args.keyring)
        else:
            return content

    smirror = simplestreams.mirrors.UrlMirrorReader(
        mirror_url, policy=policy)

    filter_list = simplestreams.filters.get_filters(
        ['datatype=image-downloads', 'ftype=disk1.img'] + args.filters
    )
    tmirror = LibvirtMirror(filter_list, verbose=args.verbose)
    tmirror.sync(smirror, initial_path)
    clean_extraneous_images()


def libvirt_pool_name_to_useful_description_string(libvirt_pool_name):
    volume_metadata = pool_metadata[libvirt_pool_name]
    filters = ' '.join('='.join((key, volume_metadata[key])) for key in USEFUL_FIELD_NAMES)
    return ' '.join([filters, '(%s)' % volume_metadata['version_name']])


def main_query(args):
    result = query(args.filters)
    useful_result = sorted(libvirt_pool_name_to_useful_description_string(r) for r in result)
    if useful_result:
        # Only print if we have results; otherwise this will print an unwanted
        # blank line
        print(*useful_result, sep="\n")


def main_purge(args):
    purge_pool()


def main(argv=None):
    # Workaround for https://bugzilla.redhat.com/show_bug.cgi?id=1063766
    # (LP: #1228231)
    def noop(*args, **kwargs): pass
    libvirt.registerErrorHandler(noop, None)

    system_arch = subprocess.check_output(
        ['dpkg', '--print-architecture']).decode().strip()
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', '-v', action='store_true')
    # Workarounds needed here: set dest and required
    # https://bugs.python.org/issue9253#msg181855
    # https://stackoverflow.com/q/22990977/478206
    # https://stackoverflow.com/a/23354355/478206
    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = True

    sync_subparser = subparsers.add_parser('sync')
    sync_subparser.set_defaults(func=main_sync)
    sync_subparser.add_argument(
        '--path', default=None,
        help='sync from index or products file in mirror'
    )
    sync_subparser.add_argument(
        '--keyring',
        help='keyring to be specified to gpg via --keyring',
        default='/usr/share/keyrings/ubuntu-cloudimage-keyring.gpg'
    )
    sync_subparser.add_argument('--source', dest='mirror_url',
        default='http://cloud-images.ubuntu.com/releases/')
    sync_subparser.add_argument('--no-authentication', action='store_true')
    sync_subparser.add_argument('filters', nargs='*', metavar='filter',
        default=["arch=%s" % system_arch])

    query_subparser = subparsers.add_parser('query')
    query_subparser.set_defaults(func=main_query)
    query_subparser.add_argument(
        'filters', nargs='*', default=[], metavar='filter')

    purge_subparser = subparsers.add_parser('purge')
    purge_subparser.set_defaults(func=main_purge)

    args = parser.parse_args(argv)
    args.func(args)


if __name__ == '__main__':
    main()



#!/usr/bin/python3

# Copyright (C) 2014-9 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

KEY_TYPES = ['rsa', 'dsa', 'ecdsa', 'ed25519']

import os
import shutil
import subprocess
import tempfile


def _keygen(key_type, private_path):
    subprocess.check_call([
        'ssh-keygen',
        '-q',
        '-f', private_path,
        '-N', '',
        '-t', key_type,
        '-C', 'root@localhost'
    ])


def read_file(path):
    with open(path, 'r') as f:
        return f.read()


def generate_ssh_host_keys():
    cloud_init_result = {}
    known_hosts_result = []
    tmp_dir = tempfile.mkdtemp(prefix='uvt-kvm.sshtmp')
    try:
        for key_type in KEY_TYPES:
            private_path = os.path.join(tmp_dir, key_type)
            _keygen(key_type, private_path)

            # ssh-keygen(1) defines that ".pub" is appended
            public_path = private_path + ".pub"

            private_ci_key = key_type + '_private'
            public_ci_key = key_type + '_public'

            private_key = read_file(private_path)
            public_key = read_file(public_path)

            cloud_init_result[private_ci_key] = private_key
            cloud_init_result[public_ci_key] = public_key

            known_hosts_result.append(public_key)
    finally:
        shutil.rmtree(tmp_dir)

    return cloud_init_result, ''.join(known_hosts_result)



# Copyright (C) 2013-9 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import codecs
import contextlib
import errno
import itertools
import json
import os
import platform
import shutil
import subprocess
import tempfile

import libvirt
from lxml import etree
from lxml.builder import E

LIBVIRT_DNSMASQ_LEASE_FILE = '/var/lib/libvirt/dnsmasq/default.leases'
LIBVIRT_DNSMASQ_STATUS_FILE = '/var/lib/libvirt/dnsmasq/virbr0.status'

# The xmlns used for custom libvirt domain xml storage
LIBVIRT_METADATA_XMLNS = 'https://launchpad.net/uvtool/libvirt/1'


def get_libvirt_pool_object(libvirt_conn, pool_name):
    try:
        pool = libvirt_conn.storagePoolLookupByName(pool_name)
    except libvirt.libvirtError:
        raise RuntimeError("Cannot find pool %s." % repr(pool_name))
    return pool


def create_volume_from_fobj(new_volume_name, fobj, image_type='raw',
        pool_name='default'):
    """Create a new libvirt volume and populate it from a file-like object."""

    compressed_fobj = tempfile.NamedTemporaryFile()
    decompressed_fobj = tempfile.NamedTemporaryFile()
    with contextlib.closing(compressed_fobj):
        with contextlib.closing(decompressed_fobj):
            shutil.copyfileobj(fobj, compressed_fobj)
            compressed_fobj.flush()
            compressed_fobj.seek(0)  # is this necessary?
            subprocess.check_call(
                [
                    'qemu-img', 'convert', '-f', image_type, '-O', image_type,
                    compressed_fobj.name, decompressed_fobj.name
                ],
                shell=False, close_fds=False)
            decompressed_fobj.seek(0)  # is this necessary?
            return create_volume_from_fobj_with_size(
                new_volume_name=new_volume_name,
                fobj=decompressed_fobj,
                fobj_size=os.fstat(decompressed_fobj.fileno()).st_size,
                image_type=image_type,
                pool_name=pool_name
            )


def platform_arch_to_image_arch(platform_arch):
    """Maps the architecture returned by the platform module
       to those used by the ubuntu images and simplestreams
    """
    try:
        return {
            'armv7l': 'armhf',
            'i686': 'i386',
            'x86_64': 'amd64',
            'ppc64le': 'ppc64el',
        }[platform_arch]
    except KeyError:
        return platform_arch


def is_emulation_target(target_arch):
    """Checks if the current host needs emulation to run target_arch

    returns True if emulation is needed and supported.
    """
    # early exit on not supported emulations
    if target_arch != 'armhf':
        return False

    local_arch = platform_arch_to_image_arch(platform.machine())
    if target_arch == 'i386':
        # 32bit works on both with full virt
        return local_arch != 'amd64' and local_arch != 'i386'
    return local_arch != target_arch


def create_volume_from_fobj_with_size(new_volume_name, fobj, fobj_size,
        image_type, pool_name):
    conn = libvirt.open('qemu:///system')
    pool = get_libvirt_pool_object(conn, pool_name)

    if image_type == 'raw':
        extra = [E.allocation(str(fobj_size)), E.capacity(str(fobj_size))]
    elif image_type == 'qcow2':
        extra = [E.capacity('0')]
    else:
        raise NotImplementedError("Unknown image type %r." % image_type)

    new_vol = E.volume(
        E.name(new_volume_name),
        E.target(E.format(type=image_type)),
        *extra
        )
    vol = pool.createXML(etree.tostring(new_vol, encoding=str), 0)

    try:
        stream = conn.newStream(0)
        vol.upload(stream, 0, fobj_size, 0)

        def handler(stream_ignored, size, opaque_ignored):
            return fobj.read(size)

        try:
            stream.sendAll(handler, None)
        except Exception as e:
            try:
                # This unexpectedly raises an exception even on a normal call,
                # so ignore it.
                stream.abort()
            except:
                pass
            raise e
        stream.finish()
    except:
        vol.delete(flags=0)
        raise

    return vol


def volume_names_in_pool(pool_name='default'):
    conn = libvirt.open('qemu:///system')
    pool = get_libvirt_pool_object(conn, pool_name)
    return pool.listVolumes()


def delete_volume_by_name(volume_name, pool_name='default'):
    conn = libvirt.open('qemu:///system')
    pool = get_libvirt_pool_object(conn, pool_name)
    volume = pool.storageVolLookupByName(volume_name)
    volume.delete(flags=0)


def have_volume_by_name(volume_name, pool_name='default'):
    conn = libvirt.open('qemu:///system')
    pool = get_libvirt_pool_object(conn, pool_name)
    try:
        volume = pool.storageVolLookupByName(volume_name)
    except libvirt.libvirtError:
        return False
    else:
        return True


def _get_all_domains(conn=None):
    if conn is None:
        conn = libvirt.open('qemu:///system')

    # libvirt in Precise doesn't seem to have a binding for
    # virConnectListAllDomains, and it seems that we must enumerate
    # defined-by-not-running and running instances separately and in different
    # ways.

    for domain_id in conn.listDomainsID():
        yield conn.lookupByID(domain_id)

    for domain_name in conn.listDefinedDomains():
        yield conn.lookupByName(domain_name)


def _domain_element_to_volume_paths(element):
    assert element.tag == 'domain'
    return (
        source.get('file')
        for source in element.xpath(
            "/domain/devices/disk[@type='file']/source[@file]"
        )
    )


def _domain_volume_paths(domain):
    volume_paths = set()

    for flags in [0, libvirt.VIR_DOMAIN_XML_INACTIVE]:
        element = etree.fromstring(domain.XMLDesc(flags))
        volume_paths.update(_domain_element_to_volume_paths(element))

    # boot items of the domain can be stored in the pool as well
    for item in ["kernel", "initrd"]:
        try:
            volume_paths.add(element.xpath("/domain/os/%s" % item)[0].text)
        except IndexError:
            pass

    return frozenset(volume_paths)


def _volume_element_to_volume_paths(element):
    assert element.tag == 'volume'
    return itertools.chain(
        (path.text for path in element.xpath('/volume/target/path')),
        (path.text for path in element.xpath('/volume/backingStore/path')),
    )


def _volume_volume_paths(volume):
    # Volumes can depend on other volumes ("backing stores"), so return all
    # paths a volume needs to function, including the top level one.
    volume_paths = set()

    element = etree.fromstring(volume.XMLDesc(0))
    volume_paths.update(_volume_element_to_volume_paths(element))

    return frozenset(volume_paths)


def _get_all_domain_volume_paths(conn=None):
    if conn is None:
        conn = libvirt.open('qemu:///system')

    all_volume_paths = set()
    for domain in _get_all_domains(conn):
        for path in _domain_volume_paths(domain):
            try:
                volume = conn.storageVolLookupByKey(path)
            except libvirt.libvirtError:
                # ignore a lookup failure, since if a volume doesn't exist,
                # it isn't reasonable to consider what backing volumes it may
                # have
                continue
            all_volume_paths.update(_volume_volume_paths(volume))

    return frozenset(all_volume_paths)


def get_all_domain_volume_names(conn=None, filter_by_dir=None):
    # Limitation: filter_by_dir must currently end in a '/' and be the
    # canonical path as libvirt returns it. Ideally I'd filter by pool instead,
    # but the libvirt API appears to not provide any method to find what pool a
    # volume is in when looked up by key.
    if conn is None:
        conn = libvirt.open('qemu:///system')

    for path in _get_all_domain_volume_paths(conn=conn):
        volume = conn.storageVolLookupByKey(path)
        if filter_by_dir and not volume.path().startswith(filter_by_dir):
            continue
        yield volume.name()


def get_domain_macs(domain_name, conn=None):
    if conn is None:
        conn = libvirt.open('qemu:///system')

    domain = conn.lookupByName(domain_name)
    xml = etree.fromstring(domain.XMLDesc(0))
    for mac in xml.xpath(
            "/domain/devices/interface[@type='network']/mac[@address]"):
        yield mac.get('address')


def dnsmasq_lease_file_mac_to_ip(lowercase_mac):
    MAC_FIELD = 1
    IP_FIELD = 2
    MIN_FIELDS = max(MAC_FIELD, IP_FIELD)

    try:
        f = codecs.open(LIBVIRT_DNSMASQ_LEASE_FILE, 'r')
    except IOError as e:
        if e.errno == errno.ENOENT:
            return None
        else:
            raise

    with contextlib.closing(f):
        for line in f:
            fields = line.split()
            if (len(fields) >= MIN_FIELDS and
                    fields[MAC_FIELD].lower() == lowercase_mac):
                return fields[IP_FIELD]
        return None


def libvirt_dnsmasq_status_file_mac_to_ip(lowercase_mac):
    try:
        f = codecs.open(LIBVIRT_DNSMASQ_STATUS_FILE , 'r')
    except IOError as e:
        if e.errno == errno.ENOENT:
            return None
        else:
            raise

    with contextlib.closing(f):
        try:
            j = json.load(f)
        except ValueError:
            return None

    for entry in j:
        if entry.get('mac-address') == lowercase_mac:
            return entry['ip-address']

    return None


def mac_to_ip(mac):
    lowercase_mac = mac.lower()
    return (
        dnsmasq_lease_file_mac_to_ip(lowercase_mac) or
        libvirt_dnsmasq_status_file_mac_to_ip(lowercase_mac)
    )


def get_domain_ssh_known_hosts(domain_name, conn=None, prefix=None):
    if conn is None:
        conn = libvirt.open('qemu:///system')

    domain = conn.lookupByName(domain_name)
    xml = etree.fromstring(domain.XMLDesc(0))
    element = xml.xpath(
        '/domain/metadata/uvt:ssh_known_hosts',
        namespaces={'uvt': LIBVIRT_METADATA_XMLNS}
    )
    if element:
        if prefix:
            return "\n".join(
                [prefix + l for l in element[0].text.splitlines()]
            )
        else:
            return element[0].text
    else:
        return None



# Copyright (C) 2013 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
from __future__ import unicode_literals

import argparse
import contextlib
import functools
import os
import socket
import sys
import time

import pyinotify

import uvtool.libvirt
from uvtool.libvirt import (
    LIBVIRT_DNSMASQ_LEASE_FILE,
    LIBVIRT_DNSMASQ_STATUS_FILE
)

SSH_PORT = 22


class ProcessEvent(pyinotify.ProcessEvent):
    def _uvtool_process_generic(self, event):
        if event.pathname in self._uvtool_watch_files:
            self._uvtool_modified = True

    process_IN_MODIFY = _uvtool_process_generic
    process_IN_MOVED_TO = _uvtool_process_generic


class LeaseModifyWaiter(object):
    def __init__(self, watch_files=None):
        if watch_files is None:
            self.watch_files = frozenset([
                LIBVIRT_DNSMASQ_LEASE_FILE,
                LIBVIRT_DNSMASQ_STATUS_FILE,
            ])
        else:
            self.watch_files = watch_files

        self.wm = pyinotify.WatchManager()
        self.process_event = ProcessEvent()
        # API-by-inheritence: must avoid namespace collision, and told
        # by upstream not to override __init__, so cannot provide parameters.
        # So we initialise here. What a hack. This is why I consider
        # API-by-inheritence in dynamic languages (where subclasses share the
        # same attribute namespace) harmful.
        self.process_event._uvtool_watch_files = self.watch_files
        self.process_event._uvtool_modified = False
        self.notifier = pyinotify.Notifier(self.wm, self.process_event)


    def start_watching(self):
        watched_dirs = set()
        for f in self.watch_files:
            if os.path.exists(f):
                self.wm.add_watch(f, pyinotify.IN_MODIFY)
            parent_path = os.path.dirname(f)
            if parent_path not in watched_dirs:
                self.wm.add_watch(parent_path, pyinotify.IN_MOVED_TO)
                watched_dirs.add(parent_path)

    def wait(self, timeout):
        # Should use CLOCK_MONOTONIC here, but it is only available since
        # Python 3.3.
        deadline = time.time() + timeout

        remaining_time = deadline - time.time()
        while remaining_time > 0:
            if self.notifier.check_events(timeout=(remaining_time*1000)):
                self.notifier.read_events()
                self.notifier.process_events()
                if self.process_event._uvtool_modified:
                    return True
            remaining_time = deadline - time.time()
        return False

    def close(self):
        self.wm.close()


def lease_has_mac(mac):
    return uvtool.libvirt.mac_to_ip(mac) is not None


def wait_for_libvirt_dnsmasq_lease(mac, timeout):
    # Shortcut check to save inotify setup
    if lease_has_mac(mac):
        return True

    timeout_time = time.time() + timeout
    waiter = LeaseModifyWaiter()
    with contextlib.closing(waiter):
        waiter.start_watching()
        # Check after we've set up a watch to avoid the race of something
        # happening between the last check and the watch starting
        if lease_has_mac(mac):
            return True
        current_time = time.time()
        while current_time < timeout_time:
            remaining_time_to_timeout = timeout_time - current_time
            waiter.wait(timeout=remaining_time_to_timeout)
            if lease_has_mac(mac):
                return True
            current_time = time.time()
        return False


def has_open_ssh_port(host, timeout=4):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    with contextlib.closing(s):
        s.settimeout(timeout)
        try:
            s.connect((host, SSH_PORT))
        except:
            return False
        else:
            return True


def poll_for_true(fn, interval, timeout):
    timeout_time = time.time() + timeout
    while time.time() < timeout_time:
        if fn():
            return True
        # This could do with a little more care to ensure that we never
        # sleep beyond timeout_time.
        time.sleep(interval)
    return False


def wait_for_open_ssh_port(host, interval, timeout):
    return poll_for_true(
        functools.partial(has_open_ssh_port, host),
        interval, timeout
    )


def main_libvirt_dnsmasq_lease(parser, args):
    if not wait_for_libvirt_dnsmasq_lease(mac=args.mac, timeout=args.timeout):
        print("cloud-wait: timed out", file=sys.stderr)
        sys.exit(1)


def main_ssh(parser, args):
    if not wait_for_open_ssh_port(args.host, args.interval, args.timeout):
        print("cloud-wait: timed out", file=sys.stderr)
        sys.exit(1)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--timeout', type=float, default=120.0)
    subparsers = parser.add_subparsers()

    libvirt_dnsmasq_lease_parser = subparsers.add_parser(
        'libvirt-dnsmasq-lease')
    libvirt_dnsmasq_lease_parser.set_defaults(func=main_libvirt_dnsmasq_lease)
    libvirt_dnsmasq_lease_parser.add_argument('mac')

    ssh_parser = subparsers.add_parser('ssh')
    ssh_parser.set_defaults(func=main_ssh)
    ssh_parser.add_argument('--interval', type=float, default=8.0)
    ssh_parser.add_argument('host')

    args = parser.parse_args()
    args.func(parser, args)


if __name__ == '__main__':
    main()



# Copyright (C) 2013 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import distutils.core
import glob

VERSION = '0.1'


distutils.core.setup(
    name="uvtool",
    description="Library and tools for using Ubuntu Cloud Images",
    version=VERSION,
    author="Robie Basak",
    license="AGPL3+",
    packages=['uvtool.libvirt'],
    scripts=glob.glob('bin/*'),
    data_files=[
        ('/usr/share/uvtool/libvirt', ['template.xml', 'remote-wait.sh',
                                       'template-aarch64.xml',
                                       'template-ppc64le.xml',
                                       'template-s390x.xml',
                                       'template-emu-armhf.xml'])
    ],
)



# Copyright (C) 2014-9 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import unittest.mock

import uvtool.libvirt.simplestreams as simplestreams

# Some tests use more recent features of mock that are not available with mock
# 0.7.2-1 as shipped with Ubuntu Precise, but for the next six months, we still
# want backports to be possible. Skip tests on builds that don't have a recent
# enough version of mock available. Ideally, this would be a test directly for
# the version of mock required, but I cannot find an implementation of
# NormalizedVersion as described in PEP-0386. As this is a temporary hack
# anyway, this will do for now, since we know that the mock version on Ubuntu
# Precise will never change.
ON_PRECISE = unittest.mock.__version__ == '0.7.2'

FAKE_VOLUME_PRODUCT_NAME = 'com.ubuntu.cloud:server:12.04:amd64'
FAKE_VOLUME_VERSION_0 = '20131119'
FAKE_VOLUME_VERSION_1 = '20131120'
ENCODED_FAKE_VOLUME_PRODUCT_NAME_0 = simplestreams._encode_libvirt_pool_name(
    FAKE_VOLUME_PRODUCT_NAME, FAKE_VOLUME_VERSION_0)
ENCODED_FAKE_VOLUME_PRODUCT_NAME_1 = simplestreams._encode_libvirt_pool_name(
    FAKE_VOLUME_PRODUCT_NAME, FAKE_VOLUME_VERSION_1)

@unittest.skipIf(ON_PRECISE, 'mock version is too old')
@unittest.mock.patch('uvtool.libvirt.simplestreams.uvtool.libvirt')
@unittest.mock.patch('uvtool.libvirt.simplestreams.pool_metadata', new={})
@unittest.mock.patch('uvtool.libvirt.simplestreams.libvirt')
class TestSimpleStreams(unittest.TestCase):
    def testSync(self, libvirt, uvtool_libvirt):
        uvtool_libvirt.have_volume_by_name.return_value = False
        uvtool_libvirt.get_all_domain_volume_names.return_value = []
        uvtool_libvirt.volume_names_in_pool.return_value = [
            ENCODED_FAKE_VOLUME_PRODUCT_NAME_0]
        simplestreams.main(
            'sync '
            '--no-authentication '
            '--source=uvtool/tests/streams/fake_stream_0 '
            '--path streams/v1/index.json '
            'release=precise arch=amd64 '
            .split()
        )
        # Check that we have mocked libvirt correctly, which means that
        # this test is working. We expect libvirt.open to have been called at
        # least once by uvtool.libvirt.simplestreams directly. This is more of
        # an assertion about the test being correct than part of the test
        # itself.
        libvirt.assert_has_calls([unittest.mock.call.open(u'qemu:///system')])

        # create_volume_from_fobj should have been called exactly once to
        # create the volume with the name that we expect
        self.assertEqual(uvtool_libvirt.create_volume_from_fobj.call_count, 1)
        self.assertEqual(
            uvtool_libvirt.create_volume_from_fobj.call_args[0][0],
            ENCODED_FAKE_VOLUME_PRODUCT_NAME_0
        )
        # Make sure the only calls to uvtool.libvirt were ones that we have
        # either whitelisted to be query-only (no side effects), or that we
        # are checking already. This makes sure, for example, that we aren't
        # deleting any volumes that we aren't expecting to delete.
        for call in uvtool_libvirt.mock_calls:
            name = call[0]
            self.assertIn(name, [
                # side effects that we have checked already
                'get_all_domain_volume_names',

                # whitelist of query functions that produce no side effects
                'create_volume_from_fobj',
                'get_libvirt_pool_object',
                'have_volume_by_name',
                'volume_names_in_pool',
            ])

    def _testResync(self, libvirt, uvtool_libvirt, old_volume_delete_expected,
            volumes_in_use=None):
        uvtool_libvirt.have_volume_by_name.side_effect = (
            lambda name, **kwargs: name == ENCODED_FAKE_VOLUME_PRODUCT_NAME_0)
        if volumes_in_use:
            uvtool_libvirt.get_all_domain_volume_names.return_value = list(
                volumes_in_use)
        else:
            uvtool_libvirt.get_all_domain_volume_names.return_value = []
        uvtool_libvirt.volume_names_in_pool.return_value = [
            ENCODED_FAKE_VOLUME_PRODUCT_NAME_0]
        simplestreams.main(
            'sync '
            '--no-authentication '
            '--source=uvtool/tests/streams/fake_stream_0 '
            '--path streams/v1/index.json '
            'release=precise arch=amd64 '
            .split()
        )
        uvtool_libvirt.reset_mock()
        uvtool_libvirt.volume_names_in_pool.return_value = [
            ENCODED_FAKE_VOLUME_PRODUCT_NAME_0]
        simplestreams.main(
            'sync '
            '--no-authentication '
            '--source=uvtool/tests/streams/fake_stream_1 '
            '--path streams/v1/index.json '
            'release=precise arch=amd64 '
            .split()
        )
        # create_volume_from_fobj should have been called exactly once to
        # create the volume with the name that we expect
        self.assertEqual(uvtool_libvirt.create_volume_from_fobj.call_count, 1)
        self.assertEqual(
            uvtool_libvirt.create_volume_from_fobj.call_args[0][0],
            ENCODED_FAKE_VOLUME_PRODUCT_NAME_1
        )

        if old_volume_delete_expected:
            # delete_volume_by_name should have been called exactly once to
            # delete the old volume that now has a new version
            self.assertEqual(
                uvtool_libvirt.delete_volume_by_name.call_count, 1)
            self.assertEqual(
                uvtool_libvirt.delete_volume_by_name.call_args[0][0],
                ENCODED_FAKE_VOLUME_PRODUCT_NAME_0
            )
        else:
            # delete_volume_by_name should not have been called at all
            self.assertEqual(
                uvtool_libvirt.delete_volume_by_name.call_count, 0)

        # Make sure the only calls to uvtool.libvirt were ones that we have
        # either whitelisted to be query-only (no side effects), or that we
        # are checking already. This makes sure, for example, that we aren't
        # deleting any volumes that we aren't expecting to delete.
        for call in uvtool_libvirt.mock_calls:
            name = call[0]
            self.assertIn(name, [
                # delete volume checked already
                'delete_volume_by_name',
                #'get_libvirt_pool_object().storageVolLookupByName().delete',

                # side effects that we have checked already
                'get_all_domain_volume_names',

                # whitelist of query functions that produce no side effects
                'create_volume_from_fobj',
                'get_libvirt_pool_object',
                'have_volume_by_name',
                'volume_names_in_pool',
                #'get_libvirt_pool_object().storageVolLookupByName',
            ])

    def testResync(self, libvirt, uvtool_libvirt):
        self._testResync(libvirt, uvtool_libvirt, True)

    def testResyncWithDomainUsingOldVolume(self, libvirt, uvtool_libvirt):
        self._testResync(
            libvirt,
            uvtool_libvirt,
            False,
            ['foo.qcow', 'foo-ds.qcow', ENCODED_FAKE_VOLUME_PRODUCT_NAME_0]
        )




# Copyright (C) 2014-9 Canonical Ltd.
# Author: Robie Basak <robie.basak@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import unittest.mock

from uvtool.libvirt.kvm import main_ssh


class TestKVM(unittest.TestCase):
    def check_ssh(self, args_hostname, args_login_name, expected_hostname,
        expected_login_name):

        parser = unittest.mock.Mock()
        args = unittest.mock.Mock()
        args.login_name = args_login_name
        args.name = args_hostname
        args.ssh_arguments = unittest.mock.sentinel.ssh_arguments
        args.insecure = True
        with unittest.mock.patch('uvtool.libvirt.kvm.ssh') as ssh_mock:
            main_ssh(parser, args)
            ssh_mock.assert_called_with(
                expected_hostname,
                expected_login_name,
                unittest.mock.sentinel.ssh_arguments,
                insecure=True,
            )

    def test_ssh_default(self):
        # "uvt-kvm ssh foo" should use user 'ubuntu'
        self.check_ssh('foo', None, 'foo', 'ubuntu')

    def test_ssh_override_with_l_option(self):
        # "uvt-kvm ssh -l bar foo" should use user 'bar'
        self.check_ssh('foo', 'bar', 'foo', 'bar')

    def test_ssh_override_with_at_sign(self):
        # "uvt-kvm ssh bar@foo" should use user 'bar'
        self.check_ssh('bar@foo', None, 'foo', 'bar')

    def test_ssh_override_and_at_in_hostname(self):
        # "uvt-kvm ssh -l baz bar@foo"
        # In this obtuse case, the hostname has an '@' in it, so this should be
        # passed through.
        self.check_ssh('bar@foo', 'baz', 'bar@foo', 'baz')
